local MAJOR_VERSION = "AceEvent-2.0"
local MINOR_VERSION = "$Revision: 3344 $"

if not AceLibrary then error(MAJOR_VERSION .. " requires AceLibrary") end
if not AceLibrary:IsNewVersion(MAJOR_VERSION, MINOR_VERSION) then return end

if not AceLibrary:HasInstance("AceOO-2.0") then error(MAJOR_VERSION .. " requires AceOO-2.0") end

local DEBUG = true

local AceOO = AceLibrary:GetInstance("AceOO-2.0")
local Mixin = AceOO.Mixin
local AceEvent = Mixin {
						"RegisterEvent",
						"UnregisterEvent",
						"UnregisterAllEvents",
						"TriggerEvent",
					   }

function AceEvent:RegisterEvent(event, method, once)
	if type(event) ~= "string" then
		error(string.format("Bad argument #2 to `RegisterEvent' (string expected, got %s)", type(event)), 2)
	end
	if not method then
		method = event
	end
	if type(method) ~= "function" and type(method) ~= "string" then
		error(string.format("Bad argument #3 to `RegisterEvent' (string or function expected, got %s)", type(method)), 2)
	end
	if type(method) == "string" and type(self[method]) ~= "function" then
		error(string.format("Cannot register event %q to method %q, it does not exist", event, method), 2)
	end
	
	if not AceEvent.registry[event] then
		AceEvent.registry[event] = {}
		AceEvent.frame:RegisterEvent(event)
	end
	
	AceEvent.registry[event][self] = method
	
	if once then
		if not AceEvent.onceRegistry then
			AceEvent.onceRegistry = {}
		end
		if not AceEvent.onceRegistry[event] then
			AceEvent.onceRegistry[event] = {}
		end
		AceEvent.onceRegistry[event][self] = true
	else
		if AceEvent.onceRegistry and AceEvent.onceRegistry[event] then
			AceEvent.onceRegistry[event][self] = nil
		end
	end
end

function AceEvent:TriggerEvent(event, a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11, a12, a13, a14, a15, a16, a17, a18, a19, a20)
	if type(event) ~= "string" then
		error(string.format("Bad argument #2 to `TriggerEvent' (string expected, got %s)", type(event)), 2)
	end
	if AceEvent.registry[event] then
		if AceEvent.onceRegistry and AceEvent.onceRegistry[event] then
			for obj in pairs(AceEvent.onceRegistry[event]) do
				local method = AceEvent.registry[event][obj]
				AceEvent.registry[event][obj] = nil
				if type(method) == "string" then
					if obj[method] then
						obj[method](obj, a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11, a12, a13, a14, a15, a16, a17, a18, a19, a20)
					end
				else -- function
					method(a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11, a12, a13, a14, a15, a16, a17, a18, a19, a20)
				end
				AceEvent.onceRegistry[event][obj] = nil
			end
		end
		for obj, method in pairs(AceEvent.registry[event]) do
			if type(method) == "string" then
				if obj[method] then
					obj[method](obj, a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11, a12, a13, a14, a15, a16, a17, a18, a19, a20)
				end
			else -- function
				method(a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11, a12, a13, a14, a15, a16, a17, a18, a19, a20)
			end
		end
	end
end

function AceEvent:UnregisterEvent(event)
	if type(event) ~= "string" then
		error(string.format("Bad argument #2 to `UnregisterEvent' (string expected, got %s)", type(event)), 2)
	end
	
	if AceEvent.registry[event] and AceEvent.registry[event][self] then
		AceEvent.registry[event][self] = nil
	elseif DEBUG then
		error("Cannot unregister an event that you are not registered with.", 2)
	end
end

function AceEvent:UnregisterAllEvents()
	for event, data in pairs(AceEvent.registry) do
		data[self] = nil
	end
end

function AceEvent:OnEmbedDisable(target)
	self.UnregisterAllEvents(target)
end

function AceEvent:activate(oldLib, oldDeactivate)
	AceLibrary(MAJOR_VERSION).super.activate(self, oldLib, oldDeactivate)
	
	if oldLib then
		self.registry = oldLib.registry
		self.onceRegistry = oldLib.onceRegistry
		self.frame = oldLib.frame
	else
		self.registry = {}
		self.frame = CreateFrame("Frame")
	end
	self.frame:SetScript("OnEvent", function()
		if event then
			self:TriggerEvent(event, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9)
		end
	end)
	
	if oldLib then
		oldDeactivate(oldLib)
	end
end

AceLibrary:Register(AceEvent, MAJOR_VERSION, MINOR_VERSION, AceEvent.activate)
AceEvent = AceLibrary(MAJOR_VERSION)
