local MAJOR_VERSION = "AceLibraryBase-2.0"
local MAJOR_VERSION_OO = "AceLibraryBaseClass-2.0"
local MINOR_VERSION = "$Revision: 3344 $"

if not AceLibrary then error(MAJOR_VERSION .. " requires AceLibrary.") end
if not AceLibrary:IsNewVersion(MAJOR_VERSION, MINOR_VERSION) and (not AceLibrary:IsNewVersion(MAJOR_VERSION_OO, MINOR_VERSION) or not AceLibrary:HasInstance("AceOO-2.0")) then return end

local AceOO = AceLibrary:HasInstance("AceOO-2.0") and AceLibrary("AceOO-2.0")

local debugstack
local _G = getfenv(0)
if type(debug) == "table" and type(debug.traceback) == "function" then
	debugstack = debug.traceback
elseif type(_G.debugstack) == "function" then
	debugstack = _G.debugstack
else
	error(MAJOR_VERSION .. " requires either debug.traceback() or debugstack()")
end

local function error(self, message, ...)
	local stack = debugstack()
	local first = string.gsub(stack, "\n.*", "")
	local file = string.gsub(first, "^\s*(.*).lua:%d+: .*", "%1")
	file = string.gsub(file, "([%(%)%.%*%+%-%[%]%?%^%$%%])", "%%%1")
	if not message then
		local _,_,second = string.find(stack, "\n(.-)\n")
		message = "error raised! " .. second
	else
		message = string.format(message, unpack(arg))
	end
	if type(self.GetLibraryVersion) == "function" and AceLibrary:HasInstance(self:GetLibraryVersion()) then
		message = string.format("%s: %s", self:GetLibraryVersion(), message)
	elseif type(self.class) == "table" and type(self.class.GetLibraryVersion) == "function" and AceLibrary:HasInstance(self.class:GetLibraryVersion()) then
		message = string.format("%s: %s", self.class:GetLibraryVersion(), message)
	end
	local i = 1
	for s in string.gfind(stack, "\n([^\n]*)") do
		i = i + 1
		if not string.find(s, file .. "%.lua:%d+:") then
			file = string.gsub(s, "^\s*(.*).lua:%d+: .*", "%1")
			file = string.gsub(file, "([%(%)%.%*%+%-%[%]%?%^%$%%])", "%%%1")
			break
		end
	end
	for s in string.gfind(stack, "\n([^\n]*)") do
		i = i + 1
		if not string.find(s, file .. "%.lua:%d+:") then
			_G.error(message, i)
			return
		end
	end
	_G.error(message, 2)
	return
end

local function assert(self, condition, message)
	if not condition then
		if not message then
			local stack = debugstack()
			local _,_,second = string.find(stack, "\n(.-)\n")
			message = "assertion failed! " .. second
		end
		error(self, message)
		return
	end
	return condition
end

local function argCheck(self, arg, num, kind, kind2, kind3, kind4)
	if tostring(type(arg)) ~= kind then
		if kind2 then
			if tostring(type(arg)) ~= kind2 then
				if kind3 then
					if tostring(type(arg)) ~= kind3 then
						if kind4 then
							if tostring(type(arg)) ~= kind4 then
								local _,_,func = string.find(debugstack(), "\n.-\n.-`(.-)'\n")
								error(self, "Bad argument #%s to `%s' (%s, %s, %s, or %s expected, got %s)", tonumber(num) or 0/0, func, kind, kind2, kind3, kind4, type(arg))
							end
						else
							local _,_,func = string.find(debugstack(), "\n.-\n.-`(.-)'\n")
							error(self, "Bad argument #%s to `%s' (%s, %s, or %s expected, got %s)", tonumber(num) or 0/0, func, kind, kind2, kind3, type(arg))
						end
					end
				else
					local _,_,func = string.find(debugstack(), "\n.-\n.-`(.-)'\n")
					error(self, "Bad argument #%s to `%s' (%s or %s expected, got %s)", tonumber(num) or 0/0, func, kind, kind2, type(arg))
				end
			end
		else
			local _,_,func = string.find(debugstack(), "\n.-\n.-`(.-)'\n")
			error(self, "Bad argument #%s to `%s' (%s expected, got %s)", tonumber(num) or 0/0, func, tostring(kind), type(arg))
		end
	end
end

if AceLibrary:IsNewVersion(MAJOR_VERSION, MINOR_VERSION) then
	local AceLibraryBase = {}
	
	function AceLibraryBase:new()
		return {
			error = error,
			assert = assert,
			argCheck = argCheck,
		}
	end
	
	setmetatable(AceLibraryBase, { __call = AceLibraryBase.new })
	
	AceLibrary:Register(AceLibraryBase, MAJOR_VERSION, MINOR_VERSION)
	AceLibraryBase = nil
end

if AceOO then
	local AceLibraryBaseClass = AceOO.Class()
	
	AceLibraryBaseClass.error = error
	AceLibraryBaseClass.assert = assert
	AceLibraryBaseClass.argCheck = argCheck
	AceLibraryBaseClass.prototype.error = error
	AceLibraryBaseClass.prototype.assert = assert
	AceLibraryBaseClass.prototype.argCheck = argCheck
	
	AceLibrary:Register(AceLibraryBaseClass, MAJOR_VERSION_OO, MINOR_VERSION)
	AceLibraryBaseClass = nil
end