local artDescriptions = {}

artDescriptions[1] = "Normal"
artDescriptions[2] = "Silver Dragon"
artDescriptions[3] = "Gold Dragon"
artDescriptions[4] = "Skull"
artDescriptions[5] = "Dead Dragon"
artDescriptions[6] = "Vines"
artDescriptions[7] = "Woodwork"
artDescriptions[8] = "Frost"
artDescriptions[9] = "Snow"
artDescriptions[10] = "Castle"
artDescriptions[11] = "Silver Dragon & Castle"
artDescriptions[12] = "Gold Dragon & Castle"

local commands = {"/kcpf", "/kcpfp", "/KC_PlayerFrame"}
local args	   = {
	type="group", 
	args = {
		showtext = { 
			name="Show Text", 
			desc="Toggles the display of experiance text.", 
			get="ShowingText", 
			set="SetShowingText", 
			type="toggle", 
			usage="value" 
		},
		art	= {
			name="Cycle Art",
			desc="Cycles the unit frame artwork",
			type="execute",
			func="CycleArt"
		}
	}
}

KC_PlayerFrame = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceHook-2.0", "AceDebug-2.0", "AceConsole-2.0", "AceDB-2.0")


KC_PlayerFrame:RegisterDB("KC_PlayerFramePlusDB")
function KC_PlayerFrame:OnInitialize(name)
	self.super.OnInitialize(self, name)
	
	self:RegisterDefaults('profile', {showText = true, artid = 1})
	self:RegisterChatCommand(commands, args)
	self:SetArt()
	
--	self:SetDebugging(true)
end


function KC_PlayerFrame:OnEnable()
	self:ShowExp()
	
	-- Makes TargetFrame Dragable
	self:HookScript(TargetFrame, "OnMouseDown", "StartMoving")
	self:HookScript(TargetFrame, "OnMouseUp", "StopMoving")

	-- Makes TargetFrame Dragable	  
	self:HookScript(PetFrame, "OnMouseDown", "StartMoving")
	self:HookScript(PetFrame, "OnMouseUp", "StopMoving")

	self:RegisterEvent("PLAYER_XP_UPDATE", "ShowExp")
	self:RegisterEvent("PLAYER_LEVEL_UP", "ShowExp")
	self:RegisterEvent("UPDATE_EXHAUSTION", "ShowExp")
	self:Debug("OnEnable has been called")
end


function KC_PlayerFrame:ShowExp()
	local currXP = UnitXP("player")
	local nextXP = UnitXPMax("player")
	local restXP = GetXPExhaustion()
	local str

	if(restXP == nil) then
		str = format("%s / %sxp %s%%", currXP, nextXP, floor((tonumber(currXP) / tonumber(nextXP)) * 100))
	else
		str = format("%s / %sxp %s%% (+%sxp) ", currXP, nextXP, floor((tonumber(currXP) / tonumber(nextXP)) * 100),(tonumber(restXP) / 2))
	end
	
	if self:ShowingText() then 
		KC_PlayerFrameExpString:SetText(str)
	else
		KC_PlayerFrameExpString:SetText("")
	end
	
	KC_PlayerFrameExpBar:SetMinMaxValues(min(0, currXP), nextXP)
	KC_PlayerFrameExpBar:SetValue(currXP)
	self:Debug("ShowExp has been called")
end


function KC_PlayerFrame:StartMoving()
	if (IsAltKeyDown()) then this:SetMovable(true); this:StartMoving(); this:IsUserPlaced(true) end
end


function KC_PlayerFrame:StopMoving()
	this:StopMovingOrSizing()
end


function KC_PlayerFrame:SetShowingText(shown)
	self.db.profile.showText = shown 
	self:ShowExp()
end

function KC_PlayerFrame:ShowingText()
	return self.db.profile.showText
end

function KC_PlayerFrame:SetArt(id)
	if not id then id = self.db.profile.artid end
	PlayerFrameTexture:SetTexture("Interface\\AddOns\\KC_PlayerFramePlus\\Art\\"..id)
	self:Debug("ArtID: %s is showing", id)
end

function KC_PlayerFrame:CycleArt()
	local hi = 12
	local id = self.db.profile.artid + 1
	if id > hi then id = 1 end
	self:SetArt(id)
	self.db.profile.artid = id
	
	self:Print("PlayerFrame artwork is now set to [%s]", artDescriptions[id])
end